      MODULE duplicate_of_driver_constants
! These definitions must be the same as frame/module_driver_constants
! and also the same as the definitions in rsl_lite.h
         INTEGER , PARAMETER :: DATA_ORDER_XYZ = 1 
         INTEGER , PARAMETER :: DATA_ORDER_YXZ = 2 
         INTEGER , PARAMETER :: DATA_ORDER_ZXY = 3 
         INTEGER , PARAMETER :: DATA_ORDER_ZYX = 4 
         INTEGER , PARAMETER :: DATA_ORDER_XZY = 5 
         INTEGER , PARAMETER :: DATA_ORDER_YZX = 6 
         INTEGER , PARAMETER :: DATA_ORDER_XY = 1
         INTEGER , PARAMETER :: DATA_ORDER_YX = 2
      END MODULE duplicate_of_driver_constants

      SUBROUTINE f_dt_build_3d (memorder, is, ie, js, je, ks, ke, ims, ime, &
     & jms, jme, kms, kme, basic_mpi_type, datatype_mpi, param)
        USE duplicate_of_driver_constants 
        IMPLICIT NONE

        INCLUDE 'mpif.h'
 
        INTEGER, INTENT(IN) :: memorder
        INTEGER, INTENT(IN) :: is, ie, js, je, ks, ke
        INTEGER, INTENT(IN) :: ims, ime, jms, jme, kms, kme
        INTEGER, INTENT(IN) :: basic_mpi_type
        INTEGER, INTENT(OUT) :: datatype_mpi
        INTEGER, INTENT(OUT) :: param(0:2)
!local variable
        INTEGER :: datatype_mpi_temp
        INTEGER :: ierr, typesize
        INTEGER(KIND=MPI_ADDRESS_KIND) :: stride

        CALL MPI_Type_size (basic_mpi_type, typesize, ierr )

        SELECT CASE ( memorder )
          CASE ( DATA_ORDER_XYZ )
            CALL mpi_type_vector( je-js+1, ie-is+1, ime-ims+1, basic_mpi_type, datatype_mpi_temp, ierr )
            stride = (ime-ims+1)*(jme-jms+1) * typesize
            CALL mpi_type_create_hvector( ke-ks+1, 1, stride, datatype_mpi_temp, datatype_mpi, ierr )
            param(0) = is 
            param(1) = js 
            param(2) = ks 
          CASE ( DATA_ORDER_YXZ )
            CALL mpi_type_vector( ie-is+1, je-js+1, jme-jms+1, basic_mpi_type, datatype_mpi_temp, ierr )
            stride = (jme-jms+1)*(ime-ims+1) * typesize
            CALL mpi_type_create_hvector( ke-ks+1, 1, stride, datatype_mpi_temp, datatype_mpi, ierr )
            param(0) = js 
            param(1) = is 
            param(2) = ks 
          CASE ( DATA_ORDER_XZY )
            CALL mpi_type_vector( ke-ks+1, ie-is+1, ime-ims+1, basic_mpi_type, datatype_mpi_temp, ierr )
            stride = (ime-ims+1)*(kme-kms+1) * typesize
            CALL mpi_type_create_hvector( je-js+1, 1, stride, datatype_mpi_temp, datatype_mpi, ierr )
            param(0) = is 
            param(1) = ks 
            param(2) = js 
          CASE ( DATA_ORDER_YZX )
            CALL mpi_type_vector( ke-ks+1, je-js+1, jme-jms+1, basic_mpi_type, datatype_mpi_temp, ierr )
            stride = (jme-jms+1)*(kme-kms+1) * typesize
            CALL mpi_type_create_hvector( ie-is+1, 1, stride, datatype_mpi_temp, datatype_mpi, ierr )
            param(0) = js 
            param(1) = ks 
            param(2) = is 
          CASE ( DATA_ORDER_ZXY )
            CALL mpi_type_vector( ie-is+1, ke-ks+1, kme-kms+1, basic_mpi_type, datatype_mpi_temp, ierr )
            stride = (kme-kms+1)*(ime-ims+1) * typesize
            CALL mpi_type_create_hvector( je-js+1, 1, stride, datatype_mpi_temp, datatype_mpi, ierr )
            param(0) = ks 
            param(1) = is 
            param(2) = js 
          CASE ( DATA_ORDER_ZYX )
            CALL mpi_type_vector( je-js+1, ke-ks+1, kme-kms+1, basic_mpi_type, datatype_mpi_temp, ierr )
            stride = (kme-kms+1)*(jme-jms+1) * typesize
            CALL mpi_type_create_hvector( ie-is+1, 1, stride, datatype_mpi_temp, datatype_mpi, ierr )
            param(0) = ks 
            param(1) = js 
            param(2) = is 
        END SELECT
        CALL mpi_type_free( datatype_mpi_temp, ierr )
      
        RETURN
      END SUBROUTINE f_dt_build_3d

      SUBROUTINE f_dt_build_2d (memorder, is, ie, js, je, ims, ime, jms, jme, basic_mpi_type, datatype_mpi, param)
        USE duplicate_of_driver_constants
        IMPLICIT NONE

        INCLUDE 'mpif.h'

        INTEGER, INTENT(IN) :: memorder
        INTEGER, INTENT(IN) :: is, ie, js, je
        INTEGER, INTENT(IN) :: ims, ime, jms, jme
        INTEGER, INTENT(IN) :: basic_mpi_type
        INTEGER, INTENT(OUT) :: datatype_mpi
        INTEGER, INTENT(OUT) :: param(0:1)
!local variable
        INTEGER :: ierr

        SELECT CASE ( memorder )
          CASE ( DATA_ORDER_XY )
            CALL mpi_type_vector( je-js+1, ie-is+1, ime-ims+1, basic_mpi_type, datatype_mpi, ierr )
            param(0) = is
            param(1) = js
          CASE ( DATA_ORDER_YX )
            CALL mpi_type_vector( ie-is+1, je-js+1, jme-jms+1, basic_mpi_type, datatype_mpi, ierr )
            param(0) = js
            param(1) = is
        END SELECT

        RETURN
      END SUBROUTINE f_dt_build_2d

      SUBROUTINE f_dt_y_2d (shw, &
     & rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p, &
     & rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p, &
     & basic_mpi_type, memorder, &
     & ids, ide, jds, jde, &
     & ims, ime, jms, jme, & 
     & ips, ipe, jps, jpe, &
     & yp, ym, datatype_2d, param_2d)
        IMPLICIT NONE

        INCLUDE 'mpif.h'

        INTEGER, INTENT(IN) :: shw, basic_mpi_type, memorder
        INTEGER, INTENT(IN) :: rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p
        INTEGER, INTENT(IN) :: rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p
        INTEGER, INTENT(IN) :: ids, ide, jds, jde
        INTEGER, INTENT(IN) :: ims, ime, jms, jme
        INTEGER, INTENT(IN) :: ips, ipe, jps, jpe
        INTEGER, INTENT(IN) :: yp, ym
        INTEGER, INTENT(OUT) :: datatype_2d(0:3), param_2d(0:1,0:3)

        INTEGER :: is, ie, js, je

        IF ( ( ips .LE. ipe ) .AND. ( jps .LE. jpe ) ) THEN 
            IF (( yp .NE. mpi_proc_null ) .AND. ( jpe .LT. jde ) ) THEN ! removed third term, since it can be fused with the second
              IF ( rsl_sendw_p .GT. 0 ) THEN 
                je = jpe - rsl_sendbeg_p + 1
                js = je - rsl_sendw_p + 1
                IF ((ips-shw) .GT. ids) THEN 
                  is = ips-shw
                ELSE 
                  is = ids
                ENDIF
                IF ((ipe+shw) .LT. ide) THEN 
                  ie = ipe+shw
                ELSE 
                  ie = ide
                ENDIF

                CALL f_dt_build_2d (memorder, is, ie, js, je, ims, ime, jms, jme, basic_mpi_type, datatype_2d(0), param_2d(0,0))
              ENDIF
              IF ( rsl_recvw_p .GT. 0 ) THEN
                js = jpe + rsl_recvbeg_p
                je = js + rsl_recvw_p - 1
                IF ((ips-shw) .GT. ids) THEN
                  is = ips-shw
                ELSE
                  is = ids
                ENDIF
                IF ((ipe+shw) .LT. ide) THEN
                  ie = ipe+shw
                ELSE
                  ie = ide
                ENDIF

                CALL f_dt_build_2d (memorder, is, ie, js, je, ims, ime, jms, jme, basic_mpi_type, datatype_2d(1), param_2d(0,1))
              ENDIF
            ENDIF
            IF (( ym .NE. mpi_proc_null ) .AND. ( jps .GT. jds ) ) THEN ! removed third term, since it can be fused with the second
              IF ( rsl_sendw_m .GT. 0 ) THEN
                js = jps + rsl_sendbeg_m - 1
                je = js + rsl_sendw_m - 1
                IF ((ips-shw) .GT. ids) THEN
                  is = ips-shw
                ELSE
                  is = ids
                ENDIF
                IF ((ipe+shw) .LT. ide) THEN
                  ie = ipe+shw
                ELSE
                  ie = ide
                ENDIF

                CALL f_dt_build_2d (memorder, is, ie, js, je, ims, ime, jms, jme, basic_mpi_type, datatype_2d(2), param_2d(0,2))
              ENDIF
              IF ( rsl_recvw_m .GT. 0 ) THEN
                je = jps - rsl_recvbeg_m
                js = je - rsl_recvw_m + 1
                IF ((ips-shw) .GT. ids) THEN
                  is = ips-shw
                ELSE
                  is = ids
                ENDIF
                IF ((ipe+shw) .LT. ide) THEN
                  ie = ipe+shw
                ELSE
                  ie = ide
                ENDIF

                CALL f_dt_build_2d (memorder, is, ie, js, je, ims, ime, jms, jme, basic_mpi_type, datatype_2d(3), param_2d(0,3))
              ENDIF
            ENDIF
        ENDIF
      
        RETURN
      END SUBROUTINE f_dt_y_2d

      SUBROUTINE f_dt_y_3d (shw, &
     & rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p, &
     & rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p, &
     & basic_mpi_type, memorder, &
     & ids, ide, jds, jde, & !, kds, kde, &     this variables aren't used (also in external/RSL_LITE/c_code.c)
     & ims, ime, jms, jme, kms, kme, & 
     & ips, ipe, jps, jpe, kps, kpe, &
     & yp, ym, datatype_3d, param_3d)
        IMPLICIT NONE

        INCLUDE 'mpif.h'

        INTEGER, INTENT(IN) :: shw, basic_mpi_type, memorder
        INTEGER, INTENT(IN) :: rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p
        INTEGER, INTENT(IN) :: rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p
        INTEGER, INTENT(IN) :: ids, ide, jds, jde !, kds, kde     this variables aren't used (also in external/RSL_LITE/c_code.c)
        INTEGER, INTENT(IN) :: ims, ime, jms, jme, kms, kme
        INTEGER, INTENT(IN) :: ips, ipe, jps, jpe, kps, kpe
        INTEGER, INTENT(IN) :: yp, ym
        INTEGER, INTENT(OUT) :: datatype_3d(0:3), param_3d(0:2,0:3)

        INTEGER :: is, ie, js, je

        IF ( ( ips .LE. ipe ) .AND. ( jps .LE. jpe ) ) THEN 
            IF (( yp .NE. mpi_proc_null ) .AND. ( jpe .LT. jde ) ) THEN ! removed third term, since it can be fused with the second
              IF ( rsl_sendw_p .GT. 0 ) THEN 
                je = jpe - rsl_sendbeg_p + 1
                js = je - rsl_sendw_p + 1
                IF ((ips-shw) .GT. ids) THEN 
                  is = ips-shw
                ELSE 
                  is = ids
                ENDIF
                IF ((ipe+shw) .LT. ide) THEN 
                  ie = ipe+shw
                ELSE 
                  ie = ide
                ENDIF

                CALL f_dt_build_3d (memorder, is, ie, js, je, kps, kpe, ims, ime, jms, jme, kms, kme, basic_mpi_type, &
                     datatype_3d(0), param_3d(0,0))
              ENDIF
              IF ( rsl_recvw_p .GT. 0 ) THEN
                js = jpe + rsl_recvbeg_p
                je = js + rsl_recvw_p - 1
                IF ((ips-shw) .GT. ids) THEN
                  is = ips-shw
                ELSE
                  is = ids
                ENDIF
                IF ((ipe+shw) .LT. ide) THEN
                  ie = ipe+shw
                ELSE
                  ie = ide
                ENDIF

                CALL f_dt_build_3d (memorder, is, ie, js, je, kps, kpe, ims, ime, jms, jme, kms, kme, basic_mpi_type, &
                     datatype_3d(1), param_3d(0,1))
              ENDIF
            ENDIF
            IF (( ym .NE. mpi_proc_null ) .AND. ( jps .GT. jds ) ) THEN ! removed third term, since it can be fused with the second
              IF ( rsl_sendw_m .GT. 0 ) THEN
                js = jps + rsl_sendbeg_m - 1
                je = js + rsl_sendw_m - 1
                IF ((ips-shw) .GT. ids) THEN
                  is = ips-shw
                ELSE
                  is = ids
                ENDIF
                IF ((ipe+shw) .LT. ide) THEN
                  ie = ipe+shw
                ELSE
                  ie = ide
                ENDIF

                CALL f_dt_build_3d (memorder, is, ie, js, je, kps, kpe, ims, ime, jms, jme, kms, kme, basic_mpi_type, &
                     datatype_3d(2), param_3d(0,2))
              ENDIF
              IF ( rsl_recvw_m .GT. 0 ) THEN
                je = jps - rsl_recvbeg_m
                js = je - rsl_recvw_m + 1
                IF ((ips-shw) .GT. ids) THEN
                  is = ips-shw
                ELSE
                  is = ids
                ENDIF
                IF ((ipe+shw) .LT. ide) THEN
                  ie = ipe+shw
                ELSE
                  ie = ide
                ENDIF

                CALL f_dt_build_3d (memorder, is, ie, js, je, kps, kpe, ims, ime, jms, jme, kms, kme, basic_mpi_type, &
                     datatype_3d(3), param_3d(0,3))
              ENDIF
            ENDIF
        ENDIF

        RETURN
      END SUBROUTINE f_dt_y_3d

      SUBROUTINE f_dt_y_4d (shw, &
     & rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p, &
     & rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p, &
     & basic_mpi_type, memorder, &
     & ids, ide, jds, jde, & !, kds, kde, &     this variables aren't used (also in external/RSL_LITE/c_code.c)
     & ims, ime, jms, jme, kms, kme, & 
     & ips, ipe, jps, jpe, kps, kpe, &
     & ls, le, &
     & yp, ym, datatype_3d, param_3d)
        IMPLICIT NONE

        INCLUDE 'mpif.h'

        INTEGER, INTENT(IN) :: shw, basic_mpi_type, memorder
        INTEGER, INTENT(IN) :: rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p
        INTEGER, INTENT(IN) :: rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p
        INTEGER, INTENT(IN) :: ids, ide, jds, jde !, kds, kde     this variables aren't used (also in external/RSL_LITE/c_code.c)
        INTEGER, INTENT(IN) :: ims, ime, jms, jme, kms, kme
        INTEGER, INTENT(IN) :: ips, ipe, jps, jpe, kps, kpe
        INTEGER, INTENT(IN) :: ls, le
        INTEGER, INTENT(IN) :: yp, ym
        INTEGER, INTENT(OUT) :: datatype_3d(0:3), param_3d(0:2,0:3)

        INTEGER :: i, ierr, typesize
        INTEGER(KIND=MPI_ADDRESS_KIND) :: stride
        INTEGER :: datatype_3d_temp(0:3) = MPI_DATATYPE_NULL

        IF (le .GE. ls) THEN
     
! get the parameter and the datatypes for the first three dimensions
          CALL f_dt_y_3d (shw, &
     &     rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p, &
     &     rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p, &
     &     basic_mpi_type, memorder, &
     &     ids, ide, jds, jde, &
     &     ims, ime, jms, jme, kms, kme, &
     &     ips, ipe, jps, jpe, kps, kpe, &
     &     yp, ym, datatype_3d_temp, param_3d)

! it can be always compute the same way, because it involves all sizes of the
! three dimensions
          CALL MPI_Type_size( basic_mpi_type, typesize, ierr )
          stride = (ime-ims+1)*(jme-jms+1)*(kme-kms+1)*typesize

          DO i = 0,3
            IF (datatype_3d_temp(i) .NE. MPI_DATATYPE_NULL ) THEN
              CALL mpi_type_create_hvector( le-ls+1, 1, stride, datatype_3d_temp(i), datatype_3d(i), ierr )
             CALL mpi_type_free( datatype_3d_temp(i), ierr )
            ENDIF
          ENDDO
        ENDIF

        RETURN
      END SUBROUTINE f_dt_y_4d

      SUBROUTINE f_dt_x_4d (shw, &
     & rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p, &
     & rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p, &
     & basic_mpi_type, memorder, &
     & ids, ide, jds, jde, & !, kds, kde, &     this variables aren't used (also in external/RSL_LITE/c_code.c)
     & ims, ime, jms, jme, kms, kme, & 
     & ips, ipe, jps, jpe, kps, kpe, &
     & ls, le, &
     & yp, ym, datatype_3d, param_3d)
        IMPLICIT NONE

        INCLUDE 'mpif.h'

        INTEGER, INTENT(IN) :: shw, basic_mpi_type, memorder
        INTEGER, INTENT(IN) :: rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p
        INTEGER, INTENT(IN) :: rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p
        INTEGER, INTENT(IN) :: ids, ide, jds, jde !, kds, kde     this variables aren't used (also in external/RSL_LITE/c_code.c)
        INTEGER, INTENT(IN) :: ims, ime, jms, jme, kms, kme
        INTEGER, INTENT(IN) :: ips, ipe, jps, jpe, kps, kpe
        INTEGER, INTENT(IN) :: ls, le
        INTEGER, INTENT(IN) :: yp, ym
        INTEGER, INTENT(OUT) :: datatype_3d(0:3), param_3d(0:2,0:3)

        INTEGER :: i, ierr, typesize
        INTEGER(KIND=MPI_ADDRESS_KIND) :: stride
        INTEGER :: datatype_3d_temp(0:3) = MPI_DATATYPE_NULL
     
        IF (le .GE. ls) THEN
! get the parameter and the datatypes for the first three dimensions
          CALL f_dt_x_3d (shw, &
     &     rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p, &
     &     rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p, &
     &     basic_mpi_type, memorder, &
     &     ids, ide, jds, jde, &
     &     ims, ime, jms, jme, kms, kme, &
     &     ips, ipe, jps, jpe, kps, kpe, &
     &     yp, ym, datatype_3d_temp, param_3d)

! it can be always compute the same way, because it involves all sizes of the
! three dimensions
          CALL MPI_Type_size( basic_mpi_type, typesize, ierr )
          stride = (ime-ims+1)*(jme-jms+1)*(kme-kms+1)*typesize

          DO i = 0,3
            IF (datatype_3d_temp(i) .NE. MPI_DATATYPE_NULL ) THEN
              CALL mpi_type_create_hvector( le-ls+1, 1, stride, datatype_3d_temp(i), datatype_3d(i), ierr )
              CALL mpi_type_free( datatype_3d_temp(i), ierr )
            ENDIF
          ENDDO
        ENDIF

        RETURN
      END SUBROUTINE f_dt_x_4d

      SUBROUTINE f_dt_x_3d (shw, &
     & rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p, &
     & rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p, &
     & basic_mpi_type, memorder, &
     & ids, ide, jds, jde, & !, kds, kde, &     this variables aren't used (also in external/RSL_LITE/c_code.c)
     & ims, ime, jms, jme, kms, kme, & 
     & ips, ipe, jps, jpe, kps, kpe, &
     & xp, xm, datatype_3d, param_3d)
        IMPLICIT NONE

        INCLUDE 'mpif.h'

        INTEGER, INTENT(IN) :: shw, basic_mpi_type, memorder
        INTEGER, INTENT(IN) :: rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p
        INTEGER, INTENT(IN) :: rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p
        INTEGER, INTENT(IN) :: ids, ide, jds, jde !, kds, kde     this variables aren't used (also in external/RSL_LITE/c_code.c)
        INTEGER, INTENT(IN) :: ims, ime, jms, jme, kms, kme
        INTEGER, INTENT(IN) :: ips, ipe, jps, jpe, kps, kpe
        INTEGER, INTENT(IN) :: xp, xm
        INTEGER, INTENT(OUT) :: datatype_3d(0:3), param_3d(0:2,0:3)

        INTEGER :: is, ie, js, je

        IF ( ( ips .LE. ipe ) .AND. ( jps .LE. jpe ) ) THEN 
            IF (( xp .NE. mpi_proc_null ) .AND. ( ipe .LT. ide ) ) THEN ! removed third term, since it can be fused with the second
              IF ( rsl_sendw_p .GT. 0 ) THEN
                ie = ipe - rsl_sendbeg_p + 1
                is = ie - rsl_sendw_p + 1 ;
                IF ((jps-shw) .GT. jds) THEN 
                  js = jps-shw
                ELSE 
                  js = jds
                ENDIF
                IF ((jpe+shw) .LT. jde) THEN 
                  je = jpe+shw
                ELSE 
                  je = jde
                ENDIF

                CALL f_dt_build_3d (memorder, is, ie, js, je, kps, kpe, ims, ime, jms, jme, kms, kme, basic_mpi_type, &
                     datatype_3d(0), param_3d(0,0))
              ENDIF
              IF ( rsl_recvw_p .GT. 0 ) THEN
                is = ipe+rsl_recvbeg_p
                ie = is + rsl_recvw_p - 1
                IF ((jps-shw) .GT. jds) THEN 
                  js = jps-shw
                ELSE 
                  js = jds
                ENDIF
                IF ((jpe+shw) .LT. jde) THEN 
                  je = jpe+shw
                ELSE 
                  je = jde
                ENDIF

                CALL f_dt_build_3d (memorder, is, ie, js, je, kps, kpe, ims, ime, jms, jme, kms, kme, basic_mpi_type, &
                     datatype_3d(1), param_3d(0,1))
              ENDIF
            ENDIF
            IF (( xm .NE. mpi_proc_null ) .AND. ( ips .GT. ids ) ) THEN ! removed third term, since it can be fused with the second
              IF ( rsl_sendw_m .GT. 0 ) THEN
                is = ips+rsl_sendbeg_m-1
                ie = is + rsl_sendw_m-1
                IF ((jps-shw) .GT. jds) THEN 
                  js = jps-shw
                ELSE 
                  js = jds
                ENDIF
                IF ((jpe+shw) .LT. jde) THEN 
                  je = jpe+shw
                ELSE 
                  je = jde
                ENDIF

                CALL f_dt_build_3d (memorder, is, ie, js, je, kps, kpe, ims, ime, jms, jme, kms, kme, basic_mpi_type, &
                     datatype_3d(2), param_3d(0,2))
              ENDIF
              IF ( rsl_recvw_m .GT. 0 ) THEN
                ie = ips-rsl_recvbeg_m
                is = ie - rsl_recvw_m + 1
                IF ((jps-shw) .GT. jds) THEN 
                  js = jps-shw
                ELSE 
                  js = jds
                ENDIF
                IF ((jpe+shw) .LT. jde) THEN 
                  je = jpe+shw
                ELSE 
                  je = jde
                ENDIF

                CALL f_dt_build_3d (memorder, is, ie, js, je, kps, kpe, ims, ime, jms, jme, kms, kme, basic_mpi_type, &
                     datatype_3d(3), param_3d(0,3))
              ENDIF
            ENDIF
        ENDIF

        RETURN
      END SUBROUTINE f_dt_x_3d

      SUBROUTINE f_dt_x_2d (shw, &
     & rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p, &
     & rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p, &
     & basic_mpi_type, memorder, &
     & ids, ide, jds, jde, & 
     & ims, ime, jms, jme, & 
     & ips, ipe, jps, jpe, &
     & xp, xm, datatype_2d, param_2d)
        IMPLICIT NONE

        INCLUDE 'mpif.h'

        INTEGER, INTENT(IN) :: shw, basic_mpi_type, memorder
        INTEGER, INTENT(IN) :: rsl_sendbeg_m, rsl_sendw_m, rsl_sendbeg_p, rsl_sendw_p
        INTEGER, INTENT(IN) :: rsl_recvbeg_m, rsl_recvw_m, rsl_recvbeg_p, rsl_recvw_p
        INTEGER, INTENT(IN) :: ids, ide, jds, jde
        INTEGER, INTENT(IN) :: ims, ime, jms, jme
        INTEGER, INTENT(IN) :: ips, ipe, jps, jpe
        INTEGER, INTENT(IN) :: xp, xm
        INTEGER, INTENT(OUT) :: datatype_2d(0:3), param_2d(0:1,0:3)

        INTEGER :: is, ie, js, je

        IF ( ( ips .LE. ipe ) .AND. ( jps .LE. jpe ) ) THEN 
            IF (( xp .NE. mpi_proc_null ) .AND. ( ipe .LT. ide ) ) THEN ! removed third term, since it can be fused with the second
              IF ( rsl_sendw_p .GT. 0 ) THEN
                ie = ipe - rsl_sendbeg_p + 1
                is = ie - rsl_sendw_p + 1 ;
                IF ((jps-shw) .GT. jds) THEN 
                  js = jps-shw
                ELSE 
                  js = jds
                ENDIF
                IF ((jpe+shw) .LT. jde) THEN 
                  je = jpe+shw
                ELSE 
                  je = jde
                ENDIF

                CALL f_dt_build_2d (memorder, is, ie, js, je, ims, ime, jms, jme, basic_mpi_type, datatype_2d(0), param_2d(0,0))
              ENDIF
              IF ( rsl_recvw_p .GT. 0 ) THEN
                is = ipe+rsl_recvbeg_p
                ie = is + rsl_recvw_p - 1
                IF ((jps-shw) .GT. jds) THEN 
                  js = jps-shw
                ELSE 
                  js = jds
                ENDIF
                IF ((jpe+shw) .LT. jde) THEN 
                  je = jpe+shw
                ELSE 
                  je = jde
                ENDIF

                CALL f_dt_build_2d (memorder, is, ie, js, je, ims, ime, jms, jme, basic_mpi_type, datatype_2d(1), param_2d(0,1))
              ENDIF
            ENDIF
            IF (( xm .NE. mpi_proc_null ) .AND. ( ips .GT. ids ) ) THEN ! removed third term, since it can be fused with the second
              IF ( rsl_sendw_m .GT. 0 ) THEN
                is = ips+rsl_sendbeg_m-1
                ie = is + rsl_sendw_m-1
                IF ((jps-shw) .GT. jds) THEN 
                  js = jps-shw
                ELSE 
                  js = jds
                ENDIF
                IF ((jpe+shw) .LT. jde) THEN 
                  je = jpe+shw
                ELSE 
                  je = jde
                ENDIF

                CALL f_dt_build_2d (memorder, is, ie, js, je, ims, ime, jms, jme, basic_mpi_type, datatype_2d(2), param_2d(0,2))
              ENDIF
              IF ( rsl_recvw_m .GT. 0 ) THEN
                ie = ips-rsl_recvbeg_m
                is = ie - rsl_recvw_m + 1
                IF ((jps-shw) .GT. jds) THEN 
                  js = jps-shw
                ELSE 
                  js = jds
                ENDIF
                IF ((jpe+shw) .LT. jde) THEN 
                  je = jpe+shw
                ELSE 
                  je = jde
                ENDIF

                CALL f_dt_build_2d (memorder, is, ie, js, je, ims, ime, jms, jme, basic_mpi_type, datatype_2d(3), param_2d(0,3))
              ENDIF
            ENDIF
        ENDIF

        RETURN
      END SUBROUTINE f_dt_x_2d

      SUBROUTINE f_dt_period_x_3d( shw, basis_mpi_type, memorder, stag, &
     & ims, ime, jms, jme, kms, kme, &
     & ips, ipe, jps, jpe, kps, kpe, &
     & datatype_3d, param_3d )

        USE module_ddt_period, ONLY: np_x, nm_x
        IMPLICIT NONE
 
        INCLUDE 'mpif.h'

        INTEGER, INTENT (IN) :: shw, basis_mpi_type, memorder, stag
        INTEGER, INTENT (IN) :: ims, ime, jms, jme, kms, kme
        INTEGER, INTENT (IN) :: ips, ipe, jps, jpe, kps, kpe
        INTEGER, INTENT (OUT) :: datatype_3d(0:3), param_3d(0:2,0:3)

        IF ( np_x .NE. MPI_PROC_NULL ) THEN
! process on right hand edge of domain
          CALL f_dt_build_3d (memorder, ipe-shw, ipe-1, jps-shw, jpe+shw, kps, kpe, ims, ime, jms, jme, kms, kme, &
               basis_mpi_type, datatype_3d(0), param_3d(0,0))
          CALL f_dt_build_3d (memorder, ipe, ipe+shw-1+stag, jps-shw, jpe+shw, kps, kpe, ims, ime, jms, jme, kms, kme, &
               basis_mpi_type, datatype_3d(1), param_3d(0,1))
        ENDIF
        IF ( nm_x .NE. MPI_PROC_NULL ) THEN
! process on left hand edge of domain
          CALL f_dt_build_3d (memorder, ips, ips+shw-1+stag, jps-shw, jpe+shw, kps, kpe, ims, ime, jms, jme, kms, kme, &
               basis_mpi_type, datatype_3d(0), param_3d(0,0))
          CALL f_dt_build_3d (memorder, ips-shw, ips-1, jps-shw, jpe+shw, kps, kpe, ims, ime, jms, jme, kms, kme, &
               basis_mpi_type, datatype_3d(1), param_3d(0,1))
        ENDIF

        RETURN
      END SUBROUTINE f_dt_period_x_3d

      SUBROUTINE f_dt_period_x_2d( shw, basis_mpi_type, memorder, stag, &
     & ims, ime, jms, jme, &
     & ips, ipe, jps, jpe, &
     & datatype_2d, param_2d )

        USE module_ddt_period, ONLY: np_x, nm_x
        IMPLICIT NONE

        INCLUDE 'mpif.h'

        INTEGER, INTENT (IN) :: shw, basis_mpi_type, memorder, stag
        INTEGER, INTENT (IN) :: ims, ime, jms, jme
        INTEGER, INTENT (IN) :: ips, ipe, jps, jpe
        INTEGER, INTENT (OUT) :: datatype_2d(0:3), param_2d(0:1,0:3)

        IF ( np_x .NE. MPI_PROC_NULL ) THEN
! process on right hand edge of domain
          CALL f_dt_build_2d (memorder, ipe-shw, ipe-1, jps-shw, jpe+shw, ims, ime, jms, jme, &
               basis_mpi_type, datatype_2d(0), param_2d(0,0))
          CALL f_dt_build_2d (memorder, ipe, ipe+shw-1+stag, jps-shw, jpe+shw, ims, ime, jms, jme, &
               basis_mpi_type, datatype_2d(1), param_2d(0,1))
        ENDIF
        IF ( nm_x .NE. MPI_PROC_NULL ) THEN
! process on left hand edge of domain
          CALL f_dt_build_2d (memorder, ips, ips+shw-1+stag, jps-shw, jpe+shw, ims, ime, jms, jme, & 
               basis_mpi_type, datatype_2d(0), param_2d(0,0))
          CALL f_dt_build_2d (memorder, ips-shw, ips-1, jps-shw, jpe+shw, ims, ime, jms, jme, &
               basis_mpi_type, datatype_2d(1), param_2d(0,1))
        ENDIF

        RETURN
      END SUBROUTINE f_dt_period_x_2d

      SUBROUTINE f_dt_period_x_4d( shw, basis_mpi_type, memorder, stag, &
     & ims, ime, jms, jme, kms, kme, &
     & ips, ipe, jps, jpe, kps, kpe, &
     & ls, le, &
     & datatype_3d, param_3d )

        IMPLICIT NONE

        INCLUDE 'mpif.h'

        INTEGER, INTENT (IN) :: shw, basis_mpi_type, memorder, stag
        INTEGER, INTENT (IN) :: ims, ime, jms, jme, kms, kme
        INTEGER, INTENT (IN) :: ips, ipe, jps, jpe, kps, kpe
        INTEGER, INTENT (IN) :: ls, le
        INTEGER, INTENT (OUT) :: datatype_3d(0:3), param_3d(0:2,0:3)

        INTEGER :: i, ierr, typesize
        INTEGER(KIND=MPI_ADDRESS_KIND) :: stride
        INTEGER :: datatype_3d_temp(0:3)

        IF (le .GE. ls) THEN
          CALL f_dt_period_x_3d( shw, basis_mpi_type, memorder, stag, &
     &    ims, ime, jms, jme, kms, kme, &
     &    ips, ipe, jps, jpe, kps, kpe, &
     &    datatype_3d_temp, param_3d )

! it can be always compute the same way, because it involves all sizes of the
! three dimensions
          CALL MPI_Type_size( basis_mpi_type, typesize, ierr )
          stride = (ime-ims+1)*(jme-jms+1)*(kme-kms+1)*typesize

! there are always two type, because this function will be only called for the
! one at the edge
          DO i = 0,1
            CALL mpi_type_create_hvector( le-ls+1, 1, stride, datatype_3d_temp(i), datatype_3d(i), ierr )
            CALL mpi_type_free( datatype_3d_temp(i), ierr )
          ENDDO
        ELSE
          datatype_3d(0:1) = MPI_DATATYPE_NULL
        ENDIF

        RETURN
      END SUBROUTINE f_dt_period_x_4d

      SUBROUTINE f_dt_period_y_4d( shw, basis_mpi_type, memorder, stag, &
     & ims, ime, jms, jme, kms, kme, &
     & ips, ipe, jps, jpe, kps, kpe, &
     & ls, le, &
     & datatype_3d, param_3d )

        IMPLICIT NONE

        INCLUDE 'mpif.h'

        INTEGER, INTENT (IN) :: shw, basis_mpi_type, memorder, stag
        INTEGER, INTENT (IN) :: ims, ime, jms, jme, kms, kme
        INTEGER, INTENT (IN) :: ips, ipe, jps, jpe, kps, kpe
        INTEGER, INTENT (IN) :: ls, le
        INTEGER, INTENT (OUT) :: datatype_3d(0:3), param_3d(0:2,0:3)

        INTEGER :: i, ierr, typesize
        INTEGER(KIND=MPI_ADDRESS_KIND) :: stride
        INTEGER :: datatype_3d_temp(0:3)

        IF (le .GE. ls) THEN 
          CALL f_dt_period_y_3d( shw, basis_mpi_type, memorder, stag, &
     &    ims, ime, jms, jme, kms, kme, &
     &    ips, ipe, jps, jpe, kps, kpe, &
     &    datatype_3d_temp, param_3d )

! it can be always compute the same way, because it involves all sizes of the
! three dimensions
          CALL MPI_Type_size( basis_mpi_type, typesize, ierr )
          stride = (ime-ims+1)*(jme-jms+1)*(kme-kms+1)*typesize

! there are always two type, because this function will be only called for the
! one at the edge
          DO i = 0,1
            CALL mpi_type_create_hvector( le-ls+1, 1, stride, datatype_3d_temp(i), datatype_3d(i), ierr )
            CALL mpi_type_free( datatype_3d_temp(i), ierr )
          ENDDO
        ELSE
          datatype_3d(0:1) = MPI_DATATYPE_NULL
        ENDIF

        RETURN
      END SUBROUTINE f_dt_period_y_4d     

      SUBROUTINE f_dt_period_y_2d( shw, basis_mpi_type, memorder, stag, &
     & ims, ime, jms, jme, &
     & ips, ipe, jps, jpe, &
     & datatype_2d, param_2d )

        USE module_ddt_period, ONLY: nm_y, np_y
        IMPLICIT NONE

        INCLUDE 'mpif.h'

        INTEGER, INTENT (IN) :: shw, basis_mpi_type, memorder, stag
        INTEGER, INTENT (IN) :: ims, ime, jms, jme
        INTEGER, INTENT (IN) :: ips, ipe, jps, jpe
        INTEGER, INTENT (OUT) :: datatype_2d(0:3), param_2d(0:1,0:3)

        IF ( np_y .NE. MPI_PROC_NULL ) THEN
! process on right hand edge of domain
          CALL f_dt_build_2d (memorder, ipe-shw, ipe+shw, jpe, jpe+shw-1+stag, ims, ime, jms, jme, &
               basis_mpi_type, datatype_2d(0), param_2d(0,0))
          CALL f_dt_build_2d (memorder, ips-shw, ipe+shw, jpe-shw, jpe-1, ims, ime, jms, jme, &
               basis_mpi_type, datatype_2d(1), param_2d(0,1))
        ENDIF
        IF ( nm_y .NE. MPI_PROC_NULL ) THEN
! process on left hand edge of domain
          CALL f_dt_build_2d (memorder, ips-shw, ipe+shw, jps-shw, jps-1, ims, ime, jms, jme, & 
               basis_mpi_type, datatype_2d(0), param_2d(0,0))
          CALL f_dt_build_2d (memorder, ips-shw, ipe+shw, jps, jps+shw-1+stag, ims, ime, jms, jme, &
               basis_mpi_type, datatype_2d(1), param_2d(0,1))
        ENDIF

        RETURN
      END SUBROUTINE f_dt_period_y_2d

      SUBROUTINE f_dt_period_y_3d( shw, basis_mpi_type, memorder, stag, &
     & ims, ime, jms, jme, kms, kme, &
     & ips, ipe, jps, jpe, kps, kpe, &
     & datatype_3d, param_3d )

        USE module_ddt_period, ONLY: nm_y, np_y
        IMPLICIT NONE

        INCLUDE 'mpif.h'

        INTEGER, INTENT (IN) :: shw, basis_mpi_type, memorder, stag
        INTEGER, INTENT (IN) :: ims, ime, jms, jme, kms, kme
        INTEGER, INTENT (IN) :: ips, ipe, jps, jpe, kps, kpe
        INTEGER, INTENT (OUT) :: datatype_3d(0:3), param_3d(0:2,0:3)

        IF ( np_y .NE. MPI_PROC_NULL ) THEN
! process on right hand edge of domain
          CALL f_dt_build_3d (memorder, ips-shw, ipe+shw, jpe, jpe+shw-1+stag, kps, kpe, ims, ime, jms, jme, kms, kme, &
               basis_mpi_type, datatype_3d(0), param_3d(0,0))
          CALL f_dt_build_3d (memorder, ips-shw, ipe+shw, jpe-shw, jpe-1, kps, kpe, ims, ime, jms, jme, kms, kme, & 
               basis_mpi_type, datatype_3d(1), param_3d(0,1))
        ENDIF
        IF ( nm_y .NE. MPI_PROC_NULL ) THEN
! process on left hand edge of domain
          CALL f_dt_build_3d (memorder, ips-shw, ipe+shw, jps-shw, jps-1, kps, kpe, ims, ime, jms, jme, kms, kme, &
               basis_mpi_type, datatype_3d(0), param_3d(0,0))
          CALL f_dt_build_3d (memorder, ips-shw, ipe+shw, jps, jps+shw-1+stag, kps, kpe, ims, ime, jms, jme, kms, kme, &
               basis_mpi_type, datatype_3d(1), param_3d(0,1))
        ENDIF

        RETURN
      END SUBROUTINE f_dt_period_y_3d
